from setuptools import setup

setup(
    name='modelling-service-test',
    version='0.1.4',
    packages=['modellingService'],
    install_requires=['bayesian-optimization'],
    author='CAPS-AN-EN',
    author_email='caps-an-en@mediaiqdigital.com',
    description='A pyspark package for modelling-service algorithms'
)
